import bz2
import io
import os
import struct
import sys
import time

from cachetools import TTLCache
from hashlib import sha256
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.serialization import load_pem_public_key
from cryptography.hazmat.primitives.asymmetric.padding import PKCS1v15
from cryptography.hazmat.primitives.hashes import SHA256
from .update_metadata_pb2 import DeltaArchiveManifest
from .update_metadata_pb2 import InstallOperation
from .update_metadata_pb2 import Signatures


def sizeof_fmt(num, suffix="B"):
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(num) < 1024.0:
            return f"{num:3.1f}{unit}{suffix}"
        num /= 1024.0
    return f"{num:.1f}Yi{suffix}"


class BlockCache(TTLCache):
    def __init__(self, maxsize, ttl, timer=time.monotonic, getsizeof=sys.getsizeof):
        super().__init__(maxsize, ttl, timer, getsizeof)

    @property
    def usage_str(self):
        return f"{self.curr_size_str}/{self.max_size_str}"

    @property
    def curr_size_str(self):
        return sizeof_fmt(self.currsize)

    @property
    def max_size_str(self):
        return sizeof_fmt(self.maxsize)


class UpdateImageException(Exception):
    pass


class UpdateImageSignatureException(UpdateImageException):
    def __init__(self, message, signed_hash, actual_hash):
        super().__init__(self, message)
        self.signed_hash = signed_hash
        self.actual_hash = actual_hash


class UpdateImage(io.RawIOBase):
    _manifest = None
    _offset = -1
    _size = 0
    _pos = 0

    def __init__(self, update_file, cache_size=500, cache_ttl=60):
        self.update_file = update_file
        self.cache_size = cache_size
        self._cache = BlockCache(
            maxsize=cache_size * 1024 * 1024,
            ttl=cache_ttl,
        )
        with open(self.update_file, "rb") as f:
            magic = f.read(4)
            if magic != b"CrAU":
                raise UpdateImageException("Wrong header")

            major = struct.unpack(">Q", f.read(8))[0]
            if major != 1:
                raise UpdateImageException("Unsupported version")

            size = struct.unpack(">Q", f.read(8))[0]
            data = f.read(size)
            self._manifest = DeltaArchiveManifest.FromString(data)
            self._offset = f.tell()

        for blob, offset, length, f in self._blobs:
            self._size += length

    def verify(self, publickey):
        _publickey = load_pem_public_key(publickey)
        with open(self.update_file, "rb") as f:
            data = f.read(self._offset + self._manifest.signatures_offset)

        actual_hash = sha256(data).digest()
        signed_hash = _publickey.recover_data_from_signature(
            self.signature,
            PKCS1v15(),
            SHA256,
        )
        if actual_hash != signed_hash:
            raise UpdateImageSignatureException(
                "Actual hash does not match signed hash", signed_hash, actual_hash
            )

    @property
    def block_size(self):
        return self._manifest.block_size

    @property
    def signature(self):
        for signature in self._signatures:
            if signature.version == 2:
                return signature.data

        return None

    @property
    def _signatures(self):
        with open(self.update_file, "rb") as f:
            f.seek(self._offset + self._manifest.signatures_offset)
            for signature in Signatures.FromString(
                f.read(self._manifest.signatures_size)
            ).signatures:
                yield signature

    @property
    def _blobs(self):
        with open(self.update_file, "rb") as f:
            for blob in self._manifest.partition_operations:
                f.seek(self._offset + blob.data_offset)
                dst_offset = blob.dst_extents[0].start_block * self.block_size
                dst_length = blob.dst_extents[0].num_blocks * self.block_size
                if blob.type not in (0, 1):
                    raise UpdateImageException(f"Unsupported type {blob.type}")

                yield blob, dst_offset, dst_length, f

        self.expire()

    def _read_blob(self, blob, blob_offset, blob_length, f):
        if blob_offset in self._cache:
            return self._cache[blob_offset]

        if blob.type not in (
            InstallOperation.Type.REPLACE,
            InstallOperation.Type.REPLACE_BZ,
        ):
            raise NotImplementedError(
                f"Error: {InstallOperation.Type.keys()[blob.type]} has not been implemented yet"
            )

        blob_data = f.read(blob.data_length)
        if sha256(blob_data).digest() != blob.data_sha256_hash:
            raise UpdateImageException("Error: Data has wrong sha256sum")

        if blob.type == InstallOperation.Type.REPLACE_BZ:
            try:
                blob_data = bz2.decompress(blob_data)

            except ValueError as err:
                raise UpdateImageException(f"Error: {err}") from err

            if blob_length - len(blob_data) < 0:
                raise UpdateImageException(
                    f"Error: Bz2 compressed data was the wrong length {len(blob_data)}"
                )

        try:
            self._cache[blob_offset] = blob_data
        except ValueError as err:
            if str(err) != "value too large":
                raise err

        return blob_data

    @property
    def cache(self):
        return self._cache

    @property
    def size(self):
        return self._size

    def expire(self):
        self._cache.expire()

    def writable(self):
        return False

    def seekable(self):
        return False

    def readable(self):
        return True

    def seek(self, offset, whence=os.SEEK_SET):
        if whence not in (os.SEEK_SET, os.SEEK_CUR, os.SEEK_END):
            raise OSError("Not supported whence")
        if whence == os.SEEK_SET and offset < 0:
            raise ValueError("offset can't be negative")
        if whence == os.SEEK_END and offset > 0:
            raise ValueError("offset can't be positive")

        if whence == os.SEEK_SET:
            self._pos = min(max(offset, 0), self._size)
        elif whence == os.SEEK_CUR:
            self._pos = min(max(self._pos + offset, 0), self._size)
        elif whence == os.SEEK_END:
            self._pos = min(max(self._pos + offset + self._size, 0), self._size)
        return self._pos

    def tell(self):
        return self._pos

    def read(self, size=-1):
        res = self.peek(size)
        self.seek(len(res), whence=os.SEEK_CUR)
        return res

    def peek(self, size=0):
        offset = self._pos
        if offset >= self._size:
            return b""

        if size <= 0 or offset + size > self._size:
            size = self._size - offset

        res = bytearray(size)
        for blob, blob_offset, blob_length, f in self._blobs:
            if offset < blob_offset:
                continue
            if offset >= blob_offset + blob_length:
                continue

            blob_data = self._read_blob(blob, blob_offset, blob_length, f)
            blob_start_offset = max(offset - blob_offset, 0)
            blob_end_offset = min(offset - blob_offset + size, blob_length)
            data = blob_data[blob_start_offset:blob_end_offset]

            assert blob_start_offset >= 0, "blob start offset is negative number"
            assert (
                blob_end_offset <= blob_length
            ), "blob end offset is larger than blob length"
            assert blob_end_offset - blob_start_offset == len(
                data
            ), "blob start and end is larger than data"

            start_offset = blob_offset + blob_start_offset - offset
            end_offset = blob_offset + blob_end_offset - offset
            res[start_offset:end_offset] = data

            assert start_offset >= 0, "start offset is negative number"
            assert start_offset < len(res), "start offset is larger than size of data"
            assert (
                end_offset < blob_offset + blob_length
            ), "end offset is larger than size of blob"
            assert end_offset - start_offset == len(
                data
            ), "size of offsets does not equal size of data"
            assert end_offset <= len(res), "end offset is larger than size of data"
            assert res[start_offset:end_offset] == data, "data does not match"

        assert len(res) == size, "size of data does not match expected size"
        return bytes(res)



import ctypes
import threading


class KillableThread(threading.Thread):
    def run(self):
        try:
            super().run()

        except SystemExit:
            pass

    def kill(self):
        if not self.is_alive():
            return

        res = ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_ulong(self.ident),
            ctypes.py_object(SystemExit),
        )

        if res == 0:
            raise ValueError(f"Invalid thread id: {self.ident}")

        if res == 1:
            return

        # "if it returns a number greater than one, you're in trouble,
        # and you should call it again with exc=NULL to revert the effect"
        ctypes.pythonapi.PyThreadState_SetAsyncExc(
            ctypes.c_ulong(self.ident),
            None,
        )
        raise SystemError("PyThreadState_SetAsyncExc failed")



import errno
import os
import queue
import sys
import threading
import time
import warnings

from pathlib import PurePosixPath

import fuse
import ext4

from .image import UpdateImage
from .image import UpdateImageSignatureException
from .threads import KillableThread


fuse.fuse_python_api = (0, 2)


class ImageException(Exception):
    pass


class FuseArgs(fuse.FuseArgs):
    def __init__(self):
        fuse.FuseArgs.__init__(self)
        self.update_file = None

    def __str__(self):
        return (
            "\n".join(
                [
                    f"< {self.update_file} on {self.mountpoint}:",
                    f"  {self.modifiers}",
                    "  -o ",
                ]
            )
            + ",\n     ".join(self._str_core())
            + " >"
        )


class FuseOptParse(fuse.FuseOptParse):
    def __init__(self, *args, **kw):
        fuse.FuseOptParse.__init__(self, *args, **kw)

    def parse_args(self, args=None, values=None):
        _opts, _args = fuse.FuseOptParse.parse_args(self, args, values)
        if _args:
            self.fuse_args.update_file = os.path.realpath(_args.pop())
        return _opts, _args


class Stat(fuse.Stat):
    def __init__(self):
        self.st_mode = 0
        self.st_ino = 0
        self.st_dev = 0
        self.st_nlink = 0
        self.st_uid = 0
        self.st_gid = 0
        self.st_size = 0
        self.st_atime = 0
        self.st_mtime = 0
        self.st_ctime = 0


class UpdateFS(fuse.Fuse):
    version = "%prog " + fuse.__version__
    fusage = "%prog update_file mountpoint [options]"
    dash_s_do = "setsingle"
    disable_path_cache = False
    cache_debug = False
    cache_size = 500
    cache_ttl = 60

    image = None
    volume = None
    inode_cache = {}
    queue = None
    exit_threads = False

    def __init__(self, *args, **kw):
        fuse.Fuse.__init__(
            self,
            *args,
            fuse_args=FuseArgs(),
            parser_class=FuseOptParse,
            **kw,
        )
        self.parser.add_option(
            mountopt="disable_path_cache",
            action="store_true",
            help="Disable path caching",
        )
        self.parser.add_option(
            mountopt="cache_debug",
            action="store_true",
            help="Debug output for path caching",
        )
        self.parser.add_option(
            mountopt="cache_size",
            default=500,
            type="int",
            help="Size in MB of memory cache for speeding up filesytem access [default: %default]",
        )
        self.parser.add_option(
            mountopt="cache_ttl",
            default=60,
            type="int",
            help="Seconds before the memory cache will evict unused chunks [default: %default]",
        )

    @property
    def update_file(self):
        return self.fuse_args.update_file

    @property
    def mountpoint(self):
        return self.fuse_args.mountpoint

    def fuse_error(self, msg):
        print(msg, file=sys.stderr)
        self.fuse_args.setmod("showhelp")
        fuse.Fuse.main(self, self.args)
        sys.exit(1)

    def main(self, args=None):
        self.args = args
        if self.fuse_args.getmod("showhelp"):
            fuse.Fuse.main(self, args)
            return

        if self.update_file is None:
            self.fuse_error("fuse: missing update_file parameter")

        if not os.path.exists(self.update_file):
            self.fuse_error(f"fuse: File does not exist {self.update_file}")

        self.image = UpdateImage(
            self.update_file,
            cache_size=self.cache_size,
            cache_ttl=self.cache_ttl,
        )
        self.volume = ext4.Volume(self.image, offset=0)
        print("Verifying signature...")
        try:
            self.image.verify(
                self.get_inode("/usr/share/update_engine/update-payload-key.pub.pem")
                .open()
                .read()
            )
            print("Signature verified")
        except UpdateImageSignatureException:
            warnings.warn("Signature doesn't match contents", RuntimeWarning)

        threads = self.start_cache_threads()
        fuse.Fuse.main(self, args)
        self.exit_threads = True
        _ = [t.kill() for t in threads]
        _ = [t.join() for t in threads]

    def start_cache_threads(self):
        thread = KillableThread(
            target=self.expire_thread,
            args=(self,),
            name="cache-ttl",
        )
        thread.start()
        return [thread]

    # Static as it's being started by threading.Thread
    @staticmethod
    def expire_thread(self):
        prev_usage = ""
        image = self.image
        while not self.exit_threads:
            if self.cache_debug:
                usage = image.cache.usage_str

            time.sleep(1)
            image.expire()
            if not self.cache_debug:
                continue

            if prev_usage == usage:
                continue

            prev_usage = usage
            print(f"[cache-ttl] {usage}")

    def get_inode(self, path):
        if self.disable_path_cache:
            inode = self.volume.inode_at(path)
            if inode is None:
                raise FileNotFoundError()

            return inode

        if path not in self.inode_cache:
            inode = self.volume.inode_at(path)
            self.inode_cache[path] = inode
            inode.verify()

        inode = self.inode_cache[path]
        if inode is None:
            raise FileNotFoundError()

        return inode

    def statfs(self):
        superblock = self.volume.superblock
        struct = fuse.StatVfs()
        struct.f_bsize = self.volume.block_size
        struct.f_frsize = self.volume.block_size
        struct.f_blocks = superblock.s_blocks_count
        struct.f_bfree = superblock.s_free_blocks_count
        struct.f_bavail = superblock.s_free_blocks_count - superblock.s_r_blocks_count
        struct.f_files = superblock.s_inodes_count
        struct.f_ffree = superblock.s_free_inodes_count
        struct.f_favail = superblock.s_free_inodes_count
        struct.f_flag = superblock.s_flags.value
        struct.f_namemax = ext4.EXT4_NAME_LEN
        return struct

    def getattr(self, path, inode=None):
        if inode is None:
            try:
                inode = self.get_inode(path)

            except FileNotFoundError:
                print(f"{path} not found")
                return -errno.ENOENT

        _stat = Stat()
        _stat.st_mode = inode.i_mode.value
        _stat.st_ino = inode.i_uid
        _stat.st_nlink = inode.i_links_count
        _stat.st_uid = inode.i_uid
        _stat.st_gid = inode.i_gid
        _stat.st_size = inode.i_size
        _stat.st_atime = inode.i_atime
        _stat.st_mtime = inode.i_mtime
        _stat.st_ctime = inode.i_ctime
        return _stat

    def open(self, path, flags):
        try:
            inode = self.get_inode(path)
            if isinstance(inode, ext4.Directory):
                return -errno.EACCES

            mode = os.O_RDONLY | os.O_WRONLY | os.O_RDWR
            if (flags & mode) != os.O_RDONLY:
                return -errno.EACCES

            return inode

        except FileNotFoundError:
            print(f"{path} not found")
            return -errno.ENOENT

    def release(self, _, __=None):
        return

    def read(self, path, size, offset, inode=None):
        if inode is None:
            try:
                inode = self.get_inode(path)

            except FileNotFoundError:
                print(f"{path} not found")
                return -errno.ENOENT

        reader = inode.open()
        reader.seek(offset, os.SEEK_SET)
        return reader.read(size)

    def readlink(self, path, inode=None):
        if inode is None:
            try:
                inode = self.get_inode(path)

            except FileNotFoundError:
                print(f"{path} not found")
                return -errno.ENOENT

        if not isinstance(inode, ext4.SymbolicLink):
            return path

        return inode.readlink().decode("utf-8")

    def getxattr(self, path, name, _):
        try:
            inode = self.get_inode(path)
            for _name, value in inode.xattrs:
                if _name == name:
                    return value

            return None

        except FileNotFoundError:
            print(f"{path} not found")
            return -errno.ENOENT

    def listxattr(self, path, _):
        try:
            inode = self.get_inode(path)
            for name, _ in inode.xattrs:
                yield name

            return None

        except FileNotFoundError:
            print(f"{path} not found")
            return -errno.ENOENT

    def opendir(self, path):
        try:
            inode = self.get_inode(path)
            if not isinstance(inode, ext4.Directory):
                return -errno.EACCES

            return inode

        except FileNotFoundError:
            print(f"{path} not found")
            return -errno.ENOENT

    def releasedir(self, _, __=None):
        return

    def readdir(self, path, _, inode=None):
        if inode is None:
            try:
                inode = self.get_inode(path)

            except FileNotFoundError:
                print(f"{path} not found")
                return -errno.ENOENT

        for dirent, _ in inode.opendir():
            yield fuse.Direntry(dirent.name_str)



from .fuse import ImageException
from .fuse import UpdateFS
from .image import UpdateImage
from .image import UpdateImageException
from .image import UpdateImageSignatureException
from .threads import KillableThread



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: update_metadata.proto
# Protobuf Python Version: 4.25.3
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x15update_metadata.proto\x12\x16\x63hromeos_update_engine\"\xe0\x02\n\x10InstallOperation\x12;\n\x04type\x18\x01 \x02(\x0e\x32-.chromeos_update_engine.InstallOperation.Type\x12\x13\n\x0b\x64\x61ta_offset\x18\x02 \x01(\r\x12\x13\n\x0b\x64\x61ta_length\x18\x03 \x01(\r\x12\x33\n\x0bsrc_extents\x18\x04 \x03(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x12\n\nsrc_length\x18\x05 \x01(\x04\x12\x33\n\x0b\x64st_extents\x18\x06 \x03(\x0b\x32\x1e.chromeos_update_engine.Extent\x12\x12\n\ndst_length\x18\x07 \x01(\x04\x12\x18\n\x10\x64\x61ta_sha256_hash\x18\x08 \x01(\x0c\"9\n\x04Type\x12\x0b\n\x07REPLACE\x10\x00\x12\x0e\n\nREPLACE_BZ\x10\x01\x12\x08\n\x04MOVE\x10\x02\x12\n\n\x06\x42SDIFF\x10\x03\"1\n\x06\x45xtent\x12\x13\n\x0bstart_block\x18\x01 \x01(\x04\x12\x12\n\nnum_blocks\x18\x02 \x01(\x04\"z\n\nSignatures\x12@\n\nsignatures\x18\x01 \x03(\x0b\x32,.chromeos_update_engine.Signatures.Signature\x1a*\n\tSignature\x12\x0f\n\x07version\x18\x01 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\")\n\x0bInstallInfo\x12\x0c\n\x04size\x18\x01 \x01(\x04\x12\x0c\n\x04hash\x18\x02 \x01(\x0c\"\x8f\x02\n\x10InstallProcedure\x12;\n\x04type\x18\x01 \x01(\x0e\x32-.chromeos_update_engine.InstallProcedure.Type\x12<\n\noperations\x18\x02 \x03(\x0b\x32(.chromeos_update_engine.InstallOperation\x12\x35\n\x08old_info\x18\x03 \x01(\x0b\x32#.chromeos_update_engine.InstallInfo\x12\x35\n\x08new_info\x18\x04 \x01(\x0b\x32#.chromeos_update_engine.InstallInfo\"\x12\n\x04Type\x12\n\n\x06KERNEL\x10\x00\"\xaf\x03\n\x14\x44\x65ltaArchiveManifest\x12\x46\n\x14partition_operations\x18\x01 \x03(\x0b\x32(.chromeos_update_engine.InstallOperation\x12\x41\n\x0fnoop_operations\x18\x02 \x03(\x0b\x32(.chromeos_update_engine.InstallOperation\x12\x18\n\nblock_size\x18\x03 \x01(\r:\x04\x34\x30\x39\x36\x12\x19\n\x11signatures_offset\x18\x04 \x01(\x04\x12\x17\n\x0fsignatures_size\x18\x05 \x01(\x04\x12?\n\x12old_partition_info\x18\x08 \x01(\x0b\x32#.chromeos_update_engine.InstallInfo\x12?\n\x12new_partition_info\x18\t \x01(\x0b\x32#.chromeos_update_engine.InstallInfo\x12<\n\nprocedures\x18\n \x03(\x0b\x32(.chromeos_update_engine.InstallProcedure')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'update_metadata_pb2', _globals)
if _descriptor._USE_C_DESCRIPTORS == False:
  DESCRIPTOR._options = None
  _globals['_INSTALLOPERATION']._serialized_start=50
  _globals['_INSTALLOPERATION']._serialized_end=402
  _globals['_INSTALLOPERATION_TYPE']._serialized_start=345
  _globals['_INSTALLOPERATION_TYPE']._serialized_end=402
  _globals['_EXTENT']._serialized_start=404
  _globals['_EXTENT']._serialized_end=453
  _globals['_SIGNATURES']._serialized_start=455
  _globals['_SIGNATURES']._serialized_end=577
  _globals['_SIGNATURES_SIGNATURE']._serialized_start=535
  _globals['_SIGNATURES_SIGNATURE']._serialized_end=577
  _globals['_INSTALLINFO']._serialized_start=579
  _globals['_INSTALLINFO']._serialized_end=620
  _globals['_INSTALLPROCEDURE']._serialized_start=623
  _globals['_INSTALLPROCEDURE']._serialized_end=894
  _globals['_INSTALLPROCEDURE_TYPE']._serialized_start=876
  _globals['_INSTALLPROCEDURE_TYPE']._serialized_end=894
  _globals['_DELTAARCHIVEMANIFEST']._serialized_start=897
  _globals['_DELTAARCHIVEMANIFEST']._serialized_end=1328
# @@protoc_insertion_point(module_scope)
